@@ -20,14 +20,15 @@ import (
 	"strings"
 	"time"
 
-	"github.com/go-kit/kit/log"
+	"github.com/go-kit/log"
 	"github.com/pkg/errors"
 	"github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
+	"github.com/scaleway/scaleway-sdk-go/scw"
+
 	"github.com/prometheus/prometheus/discovery"
 	"github.com/prometheus/prometheus/discovery/refresh"
 	"github.com/prometheus/prometheus/discovery/targetgroup"
-	"github.com/scaleway/scaleway-sdk-go/scw"
 )
 
 // metaLabelPrefix is the meta prefix used for all meta labels.
@@ -106,7 +107,7 @@ func (c SDConfig) Name() string {
 }
 
 // secretKeyForConfig returns a secret key that looks like a UUID, even if we
-// take the actuel secret from a file.
+// take the actual secret from a file.
 func (c SDConfig) secretKeyForConfig() string {
 	if c.SecretKeyFile != "" {
 		return "00000000-0000-0000-0000-000000000000"
@@ -173,8 +174,7 @@ func init() {
 
 // Discovery periodically performs Scaleway requests. It implements
 // the Discoverer interface.
-type Discovery struct {
-}
+type Discovery struct{}
 
 func NewDiscovery(conf *SDConfig, logger log.Logger) (*refresh.Discovery, error) {
 	r, err := newRefresher(conf)
